require 'lib.Component'

Player = Component:extend()

local max_speed = 550

local vertical_gap = 64
local horizontal_gap = 64

local screen_height = tonumber(sys.get_config("display.height"))
local screen_width = tonumber(sys.get_config("display.width"))

local min_y = vertical_gap
local max_y = screen_height - vertical_gap

local min_x = horizontal_gap
local max_x = screen_width - horizontal_gap

local rules_engine_id = "/rules_engine"

function Player:new(opts)
	self.dir = vmath.vector3()
	self.input = vmath.vector3()
	self.moving = false
	self.debug_mode = true
	
	Player.super.new(self, opts)
	print('Player with go_id: ' .. self.go_id)

	self.projectile_descriptor = {
		type = "Rock",
		["name"] = "rock", 
		is_singleton = false,
		pos = vmath.vector3(700, 200, 0.5)
	}
end

function Component:resolve_dependencies() 
	--self.rock = resolve_component("rock")
	--self.rock_id = self.rock.go_id
end

function Player:set_initial_sprite(opts)
	sprite.play_flipbook("#sprite", hash("player_idle"))
end

function Player:hash(opts)
	return hash("Player")
end

function Player:update_player_position(dt) 
	local pos = go.get_position()

	pos = pos + self.dir * max_speed * dt
	self:clamp_to_be_in_screen(pos)
	go.set_position(pos)
	self.projectile_descriptor.pos = pos
end

function Player:reset_movement_vectors() 
	self.input.x = 0
	self.input.y = 0
	self.dir.x, self.dir.y = 0, 0
end

function Player:on_input(go_self, action_id, action)
	self:handle_shooting_attack_keys(action_id, action)
	self:handle_melee_attack_keys(action_id, action)
	self:handle_movement_keys(action_id, action)
end

function Player:clamp_to_be_in_screen(p) 
	if p.y < min_y then
		p.y = min_y
	elseif p.y > max_y then
		p.y = max_y
	end

	if p.x < min_x then
		p.x = min_x
	elseif p.x > max_x then
		p.x = max_x
	end
end

function Player:handle_movement_keys(action_id, action) 
	if action_id == hash("key_w") or action_id == hash("key_arrow_up") then
		self.input.y = 1
		if action.released then 
			self.input.y = 0
		end
	elseif action_id == hash("key_s") or action_id == hash("key_arrow_down") then
		self.input.y = -1
		if action.released then 
			self.input.y = 0
		end
	elseif action_id == hash("key_a") or action_id == hash("key_arrow_left") then
		self.input.x = -1
		if action.released then 
			self.input.x = 0
		end
	elseif action_id == hash("key_d") or action_id == hash("key_arrow_right") then
		self.input.x = 1
		if action.released then 
			self.input.x = 0
		end
	end

	if vmath.length(self.input) > 0 then
		self.moving = true
		self.dir = vmath.normalize(self.input)
	else 
		self.moving = false
	end

	if self.debug_mode == true and action_id == hash("key_f4") and action.released then 
		msg.post(CAMERA_ID, CAMERA_SHAKE_MSG_RAW_ID, {time = CAMERA_SHAKE_SECONDS_DURATION, x = CAMERA_SHAKE_DISTANCE, y = CAMERA_SHAKE_DISTANCE, speed = 0.033})
	end
end

function Player:handle_shooting_attack_keys(action_id, action) 
	self:handle_keys(action_id, action, "key_left_control", function(self) 
		if self.debug_mode == true then
			print("Throw a rock!")
		end
		self.create_component_function(self.projectile_descriptor.type, self.projectile_descriptor.name, self.projectile_descriptor)
	end)
end

function Player:handle_melee_attack_keys(action_id, action) 
	self:handle_keys(action_id, action, "key_space", function(self) 
		if self.debug_mode == true then
			print("Hit!")
		end
	end)
end

function Player:handle_keys(action_id, action, target_key_id, callback) 
	if action_id == hash(target_key_id) and action.released then 
		callback(self)
	end
end

function Player:trigger_animation() 
	if self.moving == false then 
		self:stop_movement_animation()
	else
		self:start_movement_animation()
	end
end

function Player:stop_movement_animation(self) 
	-- TODO:
end

function Player:start_movement_animation(self) 
	-- go.set(".", "euler.z", -5)
	-- go.animate(".", "euler.z", go.PLAYBACK_LOOP_PINGPONG, 5, go.EASING_INOUTSINE, 2)
end

function Player:handle_collision(other_id) 
	if self.debug_mode == true then 
		-- print('collided with: ' .. other_id)
	end
	-- TODO: custom logic here
end

function Player:update_gun_position(dt)
	--if self.rock.thrown ~= true then 
		--go.set_position(go.get_position(), self.rock_id)
	--end
end

function Player:update(go_self, dt)
	self:update_player_position(dt)
	self:reset_movement_vectors()
	--self:update_gun_position(dt)
end

function Player:on_message(go_self, message_id, message, sender)
	Player.super.on_message(self, go_self, message_id, message, sender)
end

